// Datoteka zad2.cpp
// Simulacija ABP (Alternating Bit Protocol) protokola

#include "aes.h"
#include <iostream>
#include <cstdint>
#include <ctime>

// Ukljucujemo prostore imena aes i std
using namespace aes;
using namespace std;

// Konstante
const uint32_t velicina_paketa         = 1024; // bita
const uint32_t velicina_paketa_potvrde = 1024; // bita
const uint32_t protok_linije           = 4096; // bit/s

// Verovatnoca da ce poslati paket biti primljen (da stize bez greske)
double   prob;

// Vreme slanja paketa
const double   vreme_paket        = velicina_paketa/(1.0*protok_linije); 
// Vreme pristiznje potvrde
const double   vreme_potvrda      = velicina_paketa_potvrde/(1.0*protok_linije);
// Vreme isteka pauze (TIMEOUT)
const double   vreme_isteka_pauze = 0.52;

// Funkcija vraca slucajan broj na intervalu od 0 do 1
double random() { return rand()/(1.0*RAND_MAX); }

// Deklaracija unapred klasa
class prijemnik;
class istek_pauze;

// Klasa agenta predajnika
class predajnik: public agent {
  // Bit protokola na predaji
  uint8_t _bit;
  // Pokazivac na objekat simulacije
  simulacija* _sim;
  // Pokazivac na objekat prijemnika
  prijemnik*  _prijemnik;
  // Pokazivac na istek pauze (timeout)
  istek_pauze *_istek_pauze;
public:
  // Konstruktor
  predajnik(simulacija* s):_bit(0),_sim(s) {}
public:
  // Vrati i postavi bit
  uint8_t vrati_bit() const { return _bit; }
  void postavi_bit(const uint8_t b) { _bit = b; }
  // Vrati i postavi istek pauze  
  istek_pauze* vrati_istek_pauze() const { return _istek_pauze; }
  void postavi_istek_pauze(const istek_pauze *ip) { 
    _istek_pauze = const_cast<istek_pauze*>(ip);
  }  
  // Postavi prijemnik
  void postavi_prijemnik(const prijemnik* p) {
    _prijemnik = const_cast<prijemnik*>(p);
  }
  // Postavi prijemnik
  prijemnik* vrati_prijemnik() const {  return _prijemnik;  }
public:
  // Posalji paket
  void posalji_paket();
};

// Klasa agenta prijemnika
class prijemnik: public agent {
  // Bit protokola na prijemu
  uint8_t   _bit;
  // Broj primljenih bita
  uint32_t _broj_primljenih_bita;
  // Pokazivac na objekat simulacije
  simulacija* _sim;
  // Pokazivac na objekat predajnika
  predajnik*  _predajnik;
public:
  // Konstruktor
  prijemnik(simulacija* s):_bit(0),_broj_primljenih_bita(0),_sim(s) {}
public:
  // Vraca broj primljenih bita
  uint32_t vrati_broj_primljenih_bita() const { return _broj_primljenih_bita; }
  // Vrati i postavi bit
  uint8_t vrati_bit() const { return _bit; }
  void postavi_bit(const uint8_t b) { _bit = b; }
  // Postavi predajnika
  void postavi_predajnik(const predajnik* p) {
    _predajnik = const_cast<predajnik*>(p);
  }
  // Primi paket
  void primi_paket();
};

// Klasa dogadjaja zavrsetka simulacije
class kraj: public dogadjaj {
  // Pokazivac na objekat simulacije
  simulacija* _sim;
public:
  // Konstruktor dogadjaja slanje
  kraj(simulacija* s,vreme vn):dogadjaj(0),_sim(s) {
    postavi_vreme(_sim->vrati_vreme()+vn);
  }
public:
  // Akcija dogadjaja
  void akcija() {
    _sim->zaustavi();
  }
};

// Klasa dogadjaja slanja paketa
class slanje: public dogadjaj {
  // Pokazivac na objekat simulacije
  simulacija* _sim;
public:
  // Konstruktor dogadjaja slanje
  slanje(simulacija* s,predajnik* p,vreme vn):dogadjaj(1),_sim(s) {
    postavi_agenta(0,p);
    postavi_vreme(_sim->vrati_vreme()+vn);
  }
public:
  // Akcija dogadjaja
  void akcija();
};

// Klasa dogadjaja prijema paketa
class primi: public dogadjaj {
  // Pokazivac na objekat simulacije
  simulacija* _sim;
public:
  // Konstruktor
  primi(simulacija* s,prijemnik* p,const vreme vn):dogadjaj(1),_sim(s) {
    postavi_agenta(0,p);
    postavi_vreme(_sim->vrati_vreme()+vn);    
  }
public:
  // Akcija dogadjaja
  void akcija();
};

// Klasa dogadjaja isteka pauze
class istek_pauze: public dogadjaj {
  // Pokazivac na objekat simulacije
  simulacija* _sim;
  // Istek pauze se ne odigrava zato sto je paket potvrde stigao
  bool _pasivan;
public:
  // Konstruktor
  istek_pauze(simulacija* s,predajnik* p,const vreme vn):dogadjaj(1),_sim(s),_pasivan(false)  {
    p->postavi_istek_pauze(this);
    postavi_agenta(0,p);
    postavi_vreme(_sim->vrati_vreme()+vn);
  }
public:
  // Pasivizacija isteka pauze
  void pasivizuj() {
    _pasivan = true;
  }
  // Akcija dogadjaja
  void akcija();
};

// Klasa dogadjaja prijema potvrde
class prijem_potvrde: public dogadjaj {
  // Pokazivac na objekat simulacije
  simulacija* _sim;
public:
  // Konstruktor
  prijem_potvrde(simulacija* s,predajnik* p,const vreme vn):dogadjaj(1),_sim(s) {
    postavi_agenta(0,p);
    postavi_vreme(_sim->vrati_vreme()+vn);
  }
public:
  // Akcija dogadjaja
  void akcija();
};

// Akcija dogadjaja slanja paketa
void slanje::akcija() {
  // Predajnik
  predajnik* pred = dynamic_cast<predajnik*>(vrati_agenta(0));
  
  // Po slanju paketa rasporedjuje se dogadjaj isteka pauze (za predajnik)
  _sim->rasporedi(new istek_pauze(_sim,pred,vreme_isteka_pauze));
  cout << "Set TIMEOUT" << endl;
  
  if(random()<prob) {
    // Po slanju paketa rasporedjuje se dogadjaj prijema paketa (za prijemnik)
    _sim->rasporedi(new primi(_sim,pred->vrati_prijemnik(),vreme_paket));
    cout << "SEND(Data," << static_cast<int>(pred->vrati_bit()) << ")" << endl;
  }
}

// Akcija dogadjaja prijema paketa
void primi::akcija() {
  // Prijemnik
  prijemnik *pri = dynamic_cast<prijemnik*>(vrati_agenta(0));
  // Prijemnik prima paket
  pri->primi_paket();
}

// Akcija dogadjaja istek pauze
void istek_pauze::akcija() {
  // Predajnik
  predajnik *pred = dynamic_cast<predajnik*>(vrati_agenta(0));
  // Ukoliko istek pauze nije pasivizovan
  if(!_pasivan) {
    cout << "TIMEOUT Expired" << endl;
    // saljemo novi paket.
    pred->posalji_paket();
  }
}

// Akcija dogadjaja prijem potvrde
void prijem_potvrde::akcija() {
  // Predajnik
  predajnik *pred = dynamic_cast<predajnik*>(vrati_agenta(0));
  cout << "ACK" << endl;
  // Pasivizujemo istek pauze
  pred->vrati_istek_pauze()->pasivizuj();
  // Menjamo bit na predaji
  pred->postavi_bit(!pred->vrati_bit());
  // Saljemo novi paket;
  pred->posalji_paket();
}

// Posalji paket
void predajnik::posalji_paket() {
  // Raporedjujemo dogadjaj slanja paketa (za predajnik)
  _sim->rasporedi(new slanje(_sim,this,0.0));
}

// Primi paket
void prijemnik::primi_paket() {
  cout << "RCV(Data," << static_cast<int>(_predajnik->vrati_bit()) << ")" << endl;
  // Ukoliko se biti predjnika i prijemnika slazu
  if(_bit == _predajnik->vrati_bit()) {
    // Menjamo bit na prijemu
    _bit = !_bit;
    // Uvecavamo broj primljenih bita poruke
    _broj_primljenih_bita += velicina_paketa;
    // Rasporedi dogadjaj prijema potvrde (za predajnik)
    _sim->rasporedi(new prijem_potvrde(_sim,_predajnik,vreme_potvrda));
  }
  else {
    cout << "Dump(Data)" << endl;
  }
}

int main() {
  // Bit Error Rate
  double BER;
  // Vreme simulacije
  double vreme_sim;
  //
  srand((int)time(nullptr));
  // Unesite vreme simulacije
  cout << "Unesite vreme simulacije (s): ";
  cin >> vreme_sim;
  do {
    // Unos verovatnoce bitske greske
    cout << "Unesite BER: ";
    cin >> BER;
    // Izracunavanje verovatnoce
    prob = 1-(BER*velicina_paketa);
  } while(prob<0);
  
  // Objekat simulacije
  simulacija sim;
  // Predajnik
  predajnik tx(&sim);
  // Prijemnika
  prijemnik rx(&sim);
  
  // Uspostavljamo vezu izmedju prijemnika i predajnika
  tx.postavi_prijemnik(&rx);
  rx.postavi_predajnik(&tx);
  
  // Predajnik salje paket prijemniku
  tx.posalji_paket();
  
  // Rasporedi dogadjaj zavrsetka simulacije
  sim.rasporedi(new kraj(&sim,vreme_sim));
  
  // Zapocinjemo simulaciju
  sim.izvrsi();
  
  // Rezultat
  cout << "Protok = " << (rx.vrati_broj_primljenih_bita()/vreme_sim) << " bps" << endl;  

  return 0;
}