/*
    Agent - Event Scheduling  (AES) Library 
    Copyright (C) 2017  Marko Djogatovic

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Zaglavlje aes_sc.h
#ifndef __AES_SC_H__
#define __AES_SC_H__

#include "aes.h"

using namespace std;

namespace esa {
  // Deklaracija unapred klase dijagrama stanja  
  class dijagram_stanja;

  // Klasa stanja
  class stanje {
  protected:
    dijagram_stanja* _ds;
  public:
    stanje(const dijagram_stanja* ds):_ds(ds) {}
    virtual ~stanje() {}
  public:
    virtual void reaguj() = 0;
  };
  
  // Klasa dogadjaj dijagrama stanja
  class dogadjaj_ds {
  protected:
    dijagram_stanja* _ds;
  public:
    dogadjaj(const dijagram_stanja* ds):_ds(ds) {}
    virtual ~dogadjaj() {}
  public:
    virtual void akcija() = 0;
  };

  // Klasa dijagrama stanja
  class dijagram_stanja:public agent {
    // Indikator da je dijagram u finalnom stanju
    bool _zavrseno;
    // Tekuce stanje
    stanje *_tekuce;
  public:
    // Kontruktor
    dijagram_stanja():_zavrseno(false),_tekuce(nullptr) {}
    // Tranzicija kao reakcija na dogadjaj koja dovodi do promene stanja
    void tranzicija(const dogadjaj* d) {
      const_cast<dogadjaj*>(d)->akcija();    
      delete const_cast<dogadjaj*>(d);
    }
  public:
    // Vraca da li je dijagram stanja u zavrsnom stanju
    bool vrati_zavrseno() const          { return _zavrseno;  }
    // Postavlja dijagram stanja u zavrsno stanje
    void postavi_zavrseno(const bool f)  { _zavrseno = f;     }
    // Postavi tekuce stanje
    void postavi_tekuce(const stanje *s) {
      if(_tekuce) delete _tekuce;
      _tekuce = const_cast<stanje*>(s);
    }
    // Vrati tekuce stanje
    stanje* vrati_tekuce() const         { return _tekuce;    }
  public:
    // Dijagram stanja reaguje na tekuce stanje
    void reaguj()                        { _tekuce->reaguj(); }
  };
}

#endif   // __AES_SC_H__